﻿using Library;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;

namespace Compiler
{
    /// <summary>
    /// A fordítót reprezentálja
    /// </summary>
    public class Compiler
    {
        public const int TotalMemorySize = 1024;
        public const int GraphicMemorySize = 100;

        // Szintaktikai hibánál kivétltelt dob
        private void ThrowSyntaxError(int lineNumber)
        {
            throw new Exception($"Syntax error at line {lineNumber}");
        }

        // Átalakítja a stringet a megfelelő operandussá
        private Operand ParseOperand(string v, int lineNumber)
        {
            if (v.StartsWith("["))
            {
                string data = v.Remove(v.Length - 1, 1).Remove(0, 1);
                int value = int.Parse(data, NumberStyles.HexNumber, CultureInfo.CurrentCulture);
                return new MemoryAddressOperand(value);
            }
            else if (v.StartsWith("$"))
            {
                string data = v.Remove(0, 1);
                byte value = byte.Parse(data, NumberStyles.HexNumber, CultureInfo.CurrentCulture);
                return new ConstantOperand(value);
            }
            throw new Exception($"Syntax error at line {lineNumber}");
        }

        // Elvégzi az effektív fordítást
        private List<Operator> CompileInternal(string inputPath, string outputPath)
        {
            int programLength = 0;
            List<Operator> operators = new List<Operator>();

            var lines = File.ReadAllLines(inputPath);

            // Végigmegyünk a sorokon
            for (int i = 0; i < lines.Length; i++)
            {
                string line = lines[i];
                if (line.StartsWith("#")) // Komment sorokat kihagyjuk
                    continue;

                var data = line.Trim().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries); // Felbontjuk a sort
                if (data.Length == 1)  // Operandus nélküli műveletek
                {
                    programLength += 1;
                    switch (data[0])
                    {
                        case "WAIT":
                            operators.Add(new WaitOperator());
                            break;
                        case "END":
                            operators.Add(new EndOperator());
                            break;
                        default:
                            ThrowSyntaxError(i);
                            break;
                    }
                }
                else if (data.Length == 2) // Egyoperandusú műveletek
                {
                    programLength += 3;
                    switch (data[0])
                    {
                        case "JZ":
                            operators.Add(new JzOperator { Operand = (MemoryAddressOperand)ParseOperand(data[1], i) });
                            break;
                        case "JC":
                            operators.Add(new JcOperator { Operand = (MemoryAddressOperand)ParseOperand(data[1], i) });
                            break;
                        default:
                            ThrowSyntaxError(i);
                            break;
                    }
                }
                else if (data.Length == 3) // Kétoperandusú műveletek
                {
                    switch (data[0])
                    {
                        case "MOV":
                            var movOp1 = ParseOperand(data[1], i);
                            var movOp2 = ParseOperand(data[2], i);
                            if (movOp2.GetType() == typeof(ConstantOperand))
                            {
                                operators.Add(new MovCOperator { Op1 = movOp1, Op2 = movOp2 });
                                programLength += 4;
                            }
                            else
                            {
                                operators.Add(new MovOperator { Op1 = movOp1, Op2 = movOp2 });
                                programLength += 5;
                            }
                            break;
                        case "SUB":
                            var subOp1 = ParseOperand(data[1], i);
                            var subOp2 = ParseOperand(data[2], i);
                            if (subOp1.GetType() == typeof(ConstantOperand))
                            {
                                operators.Add(new SubCOperator { Op1 = subOp1, Op2 = subOp2 });
                                programLength += 4;
                            }
                            else
                            {
                                operators.Add(new SubOperator { Op1 = subOp1, Op2 = subOp2 });
                                programLength += 5;
                            }
                            break;
                        case "ADD":
                            var addOp1 = ParseOperand(data[1], i);
                            var addOp2 = ParseOperand(data[2], i);
                            if (addOp1.GetType() == typeof(ConstantOperand))
                            {
                                operators.Add(new AddCOperator { Op1 = addOp1, Op2 = addOp2 });
                                programLength += 4;
                            }
                            else
                            {
                                operators.Add(new AddOperator { Op1 = addOp1, Op2 = addOp2 });
                                programLength += 5;
                            }
                            break;
                        default:
                            ThrowSyntaxError(i);
                            break;
                    }
                }
                else
                    ThrowSyntaxError(i);
                if (programLength > TotalMemorySize - GraphicMemorySize)
                {
                    throw new Exception($"Memory overflow at line {i}");
                }
            }
            return operators;
        }

        /// <summary>
        /// Lefordítja a programot
        /// </summary>
        /// <param name="inputPath">Forráskódot tartalmazó fájl</param>
        /// <param name="outputPath">A bájtkódot tartalmazó fájl</param>
        public void Compile(string inputPath, string outputPath)
        {
            var program = CompileInternal(inputPath, outputPath);
            using (StreamWriter sw = new StreamWriter(outputPath))
            using (StreamWriter swTemp = new StreamWriter(outputPath + ".temp")) // Egy plus fájlba soronként kerül be a bátjkód
            {
                foreach (var @operator in program)
                {
                    sw.Write(@operator.Compile());
                    swTemp.WriteLine(@operator.Compile());
                }
            }
        }
    }
}
